//////////////////////////////////////////////////////////////////////////////////
// OpenGL Mathematics Copyright (c) 2005 G-Truc Creation (www.g-truc.net)
//////////////////////////////////////////////////////////////////////////////////
// Author  : Christophe [Groove] Riccio (contact [at] g-truc [point] net)
// Created : 2005-12-21
// Updated : 2005-12-21
// Licence : This source is under GNU LGPL licence
// File    : gtx_double.inl
//////////////////////////////////////////////////////////////////////////////////

#ifndef __gtx_double_inl__
#define __gtx_double_inl__

#include "gtx_double.h"

namespace glm
{
    ////////////////////////////////////////////////////////////////////////
    // Trigonometric Functions

    // radians
    inline double radians(double degrees)
    {
        const double pi = 3.1415926535897932384626433832795;
        return degrees * (pi / 180.0);
    }

    inline double degrees(double radians)
    {
        const double pi = 3.1415926535897932384626433832795;
        return radians * (180.0 / pi);
    }

    inline double sin(double angle)
    {
        return std::sin(angle);
    }

    inline double cos(double angle)
    {
        return std::cos(angle);
    }

    inline double tan(double angle)
    {
        return std::tan(angle);
    }

    inline double asin(double x)
    {
        return std::asin(x);
    }

    inline double acos(double x)
    {
        return std::acos(x);
    }

    inline double atan(double y, double x)
    {
        return std::atan2(y, x);
    }

    inline double atan(double x)
    {
        return std::atan(x);
    }

    ////////////////////////////////////////////////////////////////////////
    // Exponential Functions

    inline double pow(double x, double y)
    {
        return std::pow(x, y);
    }

    inline double exp(double x)
    {
        return std::exp(x);
    }

    inline double log(double x)
    {
        return std::log(x);
    }

    inline double exp2(double x)
    {
        return std::exp(0.69314718055994530941723212145818 * x);
    }

    inline double log2(double x)
    {
        return std::log(x) / 0.69314718055994530941723212145818;
    }

    inline double sqrt(double x)
    {
        return std::sqrt(x);
    }

    inline double inversesqrt(double x)
    {
        return 1.0 / std::sqrt(x);
    }

    ////////////////////////////////////////////////////////////////////////
    // Common Functions
    inline double floor(double x)
    {
        return std::floor(x);
    }

    inline double ceil(double x)
    {
        return std::ceil(x);
    }

    inline double fract(double x)
    {
        return x - std::floor(x);
    }

    inline double mod(double x, double y)
    {
        return x - y * floor(x / y);
    }

    inline double mix(double x, double y, double a)
    {
        return x * (1.0 - a) + y * a;
    }

    inline double step(double edge, double x)
    {
        return x <= edge ? 0.0 : 1.0;
    }

    inline double smoothstep(double edge0, double edge1, double x)
    {
        double tmp = glm::clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
        return tmp * tmp * (3.0 - 2.0 * tmp);
    }

    ////////////////////////////////////////////////////////////////////////
    // Geometric Functions

    inline double length(double x)
    {
        return glm::abs(x);
    }

    inline double dot(double x, double y)
    {
        return x * y;
    }

    inline double faceforward(double N, double I, double Nref)
    {
        return dot(Nref, I) < 0.0 ? N : -N;
    }

    inline double reflect(double I, double N)
    {
        return I - N * glm::dot(N, I) * 2.0;
    }

    inline double refract(double I, double N, double eta)
    {
        double dot = glm::dot(N, I);
        double k = 1.0 - eta * eta * (1.0 - dot * dot);
        if(k < 0.0)
            return 0.0;
        else
            return eta * I - (eta * dot + glm::sqrt(k)) * N;
    }

    ////////////////////////////////////////////////////////////////////////
    // Noise Functions

    inline double noise1(double x)
    {
        int iNbr = int(x + 3.0 / 2.0) * 1103515245 + 12345;
        return double(int(iNbr / 65536.0) % 32768) / 32767.0;
    }
}

#endif//__gtx_double_inl__
